import torch


class TorchCovarianceMatrix(object):

    def __init__(self, bias=False, device='cuda:0'):
        self.device = device
        self._input_dim = None  # will be set in _init_internals
        # covariance matrix, updated during the training phase
        self._cov_mtx = None
        # average, updated during the training phase
        self._avg = None
        # number of observation so far during the training phase
        self._tlen = 0

        self.bias = bias

    def _init_internals(self, x: torch.Tensor):
        """Init the internal structures.

        The reason this is not done in the constructor is that we want to be
        able to derive the input dimension and the dtype directly from the
        data this class receives.
        """
        x = x.type(torch.float64).to(device=self.device)
        # init dtype
        if len(x.shape) >1:
            dim = x.shape[1]
        else:
            dim = x.shape[0]

        self._input_dim = dim
        # init covariance matrix
        self._cov_mtx = torch.zeros((dim, dim)).type(torch.float64).to(self.device)
        # init average
        self._avg = torch.zeros((dim)).type(torch.float64).to(self.device)
        self._tlen = 0

    def update(self, x, vae):
        """Update internal structures given a batch of data
        """
        x = x.type(torch.float64).to(device=self.device)
        if self._cov_mtx is None:
            self._init_internals(x)
        # update the covariance matrix, the average and the number of
        # observations (try to do everything inplace)
        if vae:
            if x.dim() == 3: # For single layer of LSTM; TODO: Support for Multiple layers in single LSTM instance
                x = x.squeeze().t()
            elif x.dim() == 2: # Single layer of FC Linear
                x = x.t()
            else:
                x = x.unsqueeze(0).t()

        self._cov_mtx.data += torch.matmul(x.transpose(0, 1), x)
        self._avg.data += x.sum(dim=0)
        self._tlen += x.shape[0]

    def fix(self, center=True):
        """Returns the Covariance matrix"""
        # local variables
        tlen = self._tlen
        cov_mtx = self._cov_mtx
        avg = self._avg / tlen
        if center:
            avg_mtx = torch.ger(avg, avg)
            cov_mtx -= avg_mtx
        cov_mtx = cov_mtx / tlen
        return cov_mtx



"""
This file contains alternative file writers
"""
from abc import ABC, abstractmethod
from matplotlib import pyplot as plt
from shutil import make_archive
from typing import Callable, List
import pathlib
import pandas as pd
import numpy as np
import os

try:
    from tensorboardX import SummaryWriter
except ModuleNotFoundError:
    pass

STATMAP = {
    'idim': 'intrinsic-dimensionality',
    'lsat': 'saturation',
    'cov': 'covariance-matrix'
}


class AbstractWriter(ABC):

    @abstractmethod
    def add_scalar(self, name, value, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def add_scalars(self, prefix, value_dict, global_step, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def close(self):
        pass


class CompositWriter(AbstractWriter):

    def __init__(self, writers: List[AbstractWriter]):
        super(CompositWriter, self).__init__()
        self.writers = writers

    def add_scalar(self, name, value, **kwargs):
        for w in self.writers:
            w.add_scalar(name, value, **kwargs)

    def add_scalars(self, prefix, value_dict, **kwargs):
        for w in self.writers:
            w.add_scalars(prefix, value_dict, **kwargs)

    def save(self):
        for w in self.writers:
            w.save()

    def close(self):
        for w in self.writers:
            w.close()


class CSVWriter(AbstractWriter):

    def __init__(self, savepath: str, **kwargs):
        super(CSVWriter, self).__init__()
        self.value_dict = {}
        self.savepath = savepath

    def add_scalar(self, name, value, **kwargs):
        if 'covariance' in name:
            return
        if name in self.value_dict:
            self.value_dict[name].append(value)
        else:
            self.value_dict[name] = [value]
        return

    def add_scalars(self, prefix, value_dict, **kwargs):
        for name in value_dict.keys():
            self.add_scalar(name, value_dict[name])

    def save(self):
        pd.DataFrame.from_dict(self.value_dict).to_csv(self.savepath + '.csv', sep=';')

    def close(self):
        pass


class NPYWriter(AbstractWriter):

    def __init__(self, savepath: str, zip: bool = False, **kwargs):
        super(NPYWriter, self).__init__()
        self.savepath = savepath
        self.epoch_counter = {}
        self.zip = zip

    def _update_epoch_counter(self, name: str) -> int:
        if name not in self.epoch_counter:
            self.epoch_counter[name] = 0
        else:
            self.epoch_counter[name] += 1
        return self.epoch_counter[name]

    def _get_and_create_savepath(self, name: str, epoch) -> str:
        savepath = os.path.join(self.savepath, name)
        if not os.path.exists(savepath):
            os.makedirs(savepath)
        return os.path.join(savepath, name + f'_epoch{epoch}')

    def add_scalar(self, name, value, **kwargs):
        epoch = self._update_epoch_counter(name)
        svpth = self._get_and_create_savepath(name, epoch)
        np.save(svpth, value, **kwargs)

    def add_scalars(self, prefix, value_dict, **kwargs):
        for key in value_dict.keys():
            self.add_scalar(prefix + '_' + key, value_dict[key])

    def save(self):
        if self.zip:
            make_archive(
                base_name=os.path.basename(self.savepath),
                format='zip',
                root_dir=os.path.dirname(self.savepath),
                verbose=True
            )

    def close(self):
        pass


class PrintWriter(AbstractWriter):

    def __init__(self, savepath: str, **kwargs):
        super(PrintWriter, self).__init__()

    def add_scalar(self, name, value, **kwargs):
        print(name, ':', value)

    def add_scalars(self, prefix, value_dict, **kwargs):
        for key in value_dict.keys():
            self.add_scalar(prefix + '_' + key, value_dict[key])

    def save(self):
        pass

    def close(self):
        pass


class TensorBoardWriter(AbstractWriter):

    def __init__(self, savepath: str, **kwargs):
        super(TensorBoardWriter, self).__init__()
        self.savepath = savepath
        self.writer = SummaryWriter(savepath)

    def add_scalar(self, name, value, **kwargs):
        if 'covariance' in name:
            return
        self.writer.add_scalar(name, value)

    def add_scalars(self, prefix, value_dict, **kwargs):
        self.writer.add_scalars(prefix, value_dict)

    def save(self):
        pass

    def close(self):
        self.writer.close()


class CSVandPlottingWriter(CSVWriter):

    def __init__(self, savepath: str, plot_manipulation_func: Callable[[plt.Axes], plt.Axes] = None, **kwargs):
        """
        This writer produces the CSV similar to the CSV writer, but also provides plots, which are per default stored in
        the subfolder.
        :param savepath: path to store images and csvs
        :param plot_manipulation_func: this is a function mapping an axis object to an axis object, you can use this to
                                       to manipulate the plot even further, by writing arbitrary pyplot code in this
                                       function
        :param kwargs:
        """
        super(CSVandPlottingWriter, self).__init__(savepath)
        self.plot_man_func = plot_manipulation_func if plot_manipulation_func is not None else lambda x: x
        self.primary_metric = None if not 'primary_metric' in kwargs else kwargs['primary_metric']
        self.fontsize = 16 if not 'fontsize' in kwargs else kwargs['fontsize']
        self.figsize = None if not 'figsize' in kwargs else kwargs['figsize']
        self.epoch_counter: int = 0
        self.stats = []

    def _look_for_stats(self):
        if len(self.stats) == 0:
            sat = False
            idim = False
            for key in self.value_dict.keys():
                if 'saturation' in key:
                    sat = True
                if 'intrinsic-dimensionality' in key:
                    idim = True
            if sat:
                self.stats.append('lsat')
            if idim:
                self.stats.append('idim')

    def add_scalar(self, name, value, **kwargs):
        if 'covariance' in name:
            return
        if name in self.value_dict:
            self.value_dict[name].append(value)
        else:
            self.value_dict[name] = [value]
        return

    def add_scalars(self, prefix, value_dict, **kwargs):
        for name in value_dict.keys():
            self.add_scalar(name, value_dict[name])

    def save(self):
        self._look_for_stats()
        pd.DataFrame.from_dict(self.value_dict).to_csv(self.savepath + '.csv', sep=';')
        for stat in self.stats:
            plot_stat_level_from_results(self.savepath + '.csv', stat=stat, epoch=self.epoch_counter,
                                         primary_metric=self.primary_metric, fontsize=self.fontsize,
                                         figsize=self.figsize)
        self.epoch_counter += 1

    def close(self):
        pass


def extract_layer_stat(df, epoch=19, primary_metric=None, stat='saturation'):
    cols = list(df.columns)
    train_cols = [col for col in cols if
                  'train' in col and not 'accuracy' in col and stat in col]
    epoch_df = df[df.index.values == epoch]
    pm = None if primary_metric is None else epoch_df[primary_metric].values[0]
    epoch_df = epoch_df[train_cols]
    return epoch_df, pm


def plot_stat(df, stat, pm=-1, savepath='run.png', epoch=0, primary_metric=None, fontsize=16, figsize=None,
              line=True, scatter=True, ylim=(0, 1.0), alpha_line=.6, alpha_scatter=1.0, color_line=None,
              color_scatter=None,
              primary_metric_loc=(0.7, 0.8), show_col_label_x=True, show_col_label_y=True, show_grid=True, save=True):
    plt.clf()
    if figsize is not None:
        print(figsize)
        plt.figure(figsize=figsize)
    ax = plt.gca()
    col_names = [i for i in df.columns]
    if line:
        ax.plot(list(range(len(col_names))), df.values[0], alpha=alpha_line, color=color_line)
    if scatter:
        ax.scatter(list(range(len(col_names))), df.values[0], alpha=alpha_scatter, color=color_scatter)
    plt.xticks(list(range(len(col_names))), [col_name.split('_', )[1] for col_name in col_names],
               rotation=90)
    if not ylim is None:
        ax.set_ylim(ylim)
    if primary_metric is not None:
        ax.text(primary_metric_loc[0], primary_metric_loc[1], f'{primary_metric}: {pm}')
    plt.yticks(fontsize=fontsize)
    if show_col_label_x:
        plt.xlabel('layers', fontsize=fontsize)
    plt.title(pathlib.Path(savepath).name.replace('_', ' ').replace('.csv', f' epoch: {epoch}'), fontsize=fontsize)
    if show_col_label_y:
        plt.ylabel(stat if not stat in STATMAP else STATMAP[stat], rotation='vertical', fontsize=fontsize)
    if show_grid:
        plt.grid()
    plt.tight_layout()
    if save:
        final_savepath = savepath.replace('.csv', f'{stat}_epoch_{epoch}.png')
        print(final_savepath)
        plt.savefig(final_savepath)
    return ax


def plot_stat_level_from_results(savepath, epoch, stat, primary_metric=None, fontsize=16, figsize=None, line=True,
                                 scatter=True, ylim=(0, 1.0), alpha_line=.6, alpha_scatter=1.0, color_line=None,
                                 color_scatter=None,
                                 primary_metric_loc=(0.7, 0.8), show_col_label_x=True, show_col_label_y=True,
                                 show_grid=True, save=True):
    df = pd.read_csv(savepath, sep=';')
    epoch_df, pm = extract_layer_stat(df, stat=STATMAP[stat], epoch=epoch, primary_metric=primary_metric)
    ax = plot_stat(df=epoch_df, pm=pm, savepath=savepath, epoch=epoch, primary_metric=primary_metric, fontsize=fontsize,
                   figsize=figsize, stat=stat, ylim=None if stat is 'idim' else (0, 1.0), line=line, scatter=scatter,
                   alpha_line=alpha_line, alpha_scatter=alpha_scatter, color_line=color_line,
                   color_scatter=color_scatter,
                   primary_metric_loc=primary_metric_loc, show_col_label_x=show_col_label_x,
                   show_col_label_y=show_col_label_y,
                   show_grid=show_grid, save=save)
    return ax



from typing import Optional, List, Dict, Tuple
import pandas as pd
import numpy as np
import warnings
from delve.writers import STATMAP
from delve.metrics import compute_intrinsic_dimensionality, compute_saturation
from os.path import join, basename, dirname, curdir, splitext
from pathlib import Path
import torch


def _get_files(path: str, filext: str) -> List[str]:
    paths: List[str] = [str(x) for x in Path(path).rglob(f'*.{filext}')]
    return paths


def _check_stat_available(paths: List[str], stat: str) -> bool:
    return any([stat in basename(path) for path in paths])


def _filter_by_stat(paths: List[str], stat: str, neg: bool = False) -> List[str]:
    result = [path for path in paths if stat in basename(path)]
    if neg:
        for res in result:
            paths.remove(res)
        result = paths
    return result


def _filter_by_stat_shortcuts(paths: List[str], stats: List[str], neg: bool = False) -> List[str]:
    result = []
    for stat in stats:
        result += _filter_by_stat(paths, STATMAP[stat], neg)
    return list(set(result))


def _sort_files_by_epoch(npy_files: List[str]) -> List[str]:
    npy_files.sort(key=lambda x: int(splitext(basename(x))[0].split('epoch')[-1]))
    return npy_files


def _obtain_stat_from_npy(npy_files: List[str], stat: str) -> pd.DataFrame:
    filtered_paths = _filter_by_stat(npy_files, stat)
    sorted_paths = _sort_files_by_epoch(filtered_paths)
    res = [np.load(x) for x in sorted_paths]
    return res


def _group_by_directory(npy_files: List[str]) -> Dict[str, List[str]]:
    result = {}
    for file in npy_files:
        if basename(dirname(file)) in result:
            result[basename(dirname(file))].append(file)
        else:
            result[basename(dirname(file))] = [file]
    return result


def _obtain_stats_from_npy(npy_files: List[str]) -> pd.DataFrame:
    filtered = npy_files
    grouped_files = _group_by_directory(filtered)
    df_dict = {stat: _obtain_stat_from_npy(files, stat) for stat, files in grouped_files.items()}
    return pd.DataFrame.from_dict(df_dict)


def _recompute_value_from_cov(cov_path: str, stat: str, thresh: float) -> float:
    cov = torch.from_numpy(np.load(cov_path))
    if stat == 'lsat':
        return compute_saturation(cov, thresh)
    elif stat == 'idim':
        return compute_intrinsic_dimensionality(cov, thresh)


def _recompute_stat_from_cov(cov_name, cov_files: List[str], stat: str, thresh: float) -> Dict[str, List[float]]:
    sorted_cov_files = _sort_files_by_epoch(cov_files)
    name = cov_name.replace('covariance-matrix', STATMAP[stat])
    values = []
    for cov_file in sorted_cov_files:
        value = _recompute_value_from_cov(cov_file, stat, thresh)
        values.append(value)
    return {name: values}


def _recompute_stats(npy_files: List[str], stats: List[str], thresh: float) -> pd.DataFrame:
    # get everything except covariance matrix and the stats to compute
    filtered = _filter_by_stat_shortcuts(npy_files, stats=stats, neg=True)
    filtered = _filter_by_stat_shortcuts(filtered, stats=['cov'], neg=True)
    grouped_files = _group_by_directory(filtered)

    # get the covariance matrices
    cov_files = _filter_by_stat_shortcuts(npy_files, stats=['cov'])

    # obtain stats not recomputable through the covariance matrix
    df_dict = {stat: _obtain_stat_from_npy(files, stat) for stat, files in grouped_files.items()}

    # group the covariance matrix and recompute all stats for all stats and all layers
    grouped_covariance = _group_by_directory(cov_files)
    for cov_name, cov_paths in grouped_covariance.items():
        for stat in stats:
            stat_dict = _recompute_stat_from_cov(cov_name, cov_paths, stat, thresh)
            df_dict.update(stat_dict)
    return pd.DataFrame.from_dict(df_dict)


def reconstruct_csv_from_npy_data(npywriter_out_path: str,
                                  savefile: Optional[str] = None,
                                  thresh: Optional[float] = None,
                                  stats: List[str] = ['lsat', 'idim'],
                                  ) -> pd.DataFrame:
    """
    This function allows the user to reconstruct the csv as constructed by the csv-writer.
    It further allows computing metrics that were initially not computed during training, if the covariance matrix is
    stored during training using the npy-writer.
    It also allows to recompute stats using different thresholds than originally used during training.
    :param npywriter_out_path: The path were the npy-writer has written the output
    :param savefile: the target-file to store the csv into, if not set, the data is not saved
    :param thresh: the threshold for computing the intrinsic dimensionality and saturation. If not set, the function
    tries to find stored version of all metrics. If set to any value, the stats will be computed from the stored covariance matrix.
    If not covariance matrix is given and threshold is set an error is raised.
    :param stats: the statistic to compute or read. Valid values are 'sat' and 'idim'. All other metrics that were recorded
    during training and properly stored as npy-files will be read as well.
    :return:
    """
    npy_files: List[str] = _get_files(npywriter_out_path, 'npy')
    if thresh is None:
        if 'sat' in stats and not _check_stat_available(npy_files, 'saturation'):
            warnings.warn('No .npy file was found for saturation, stat will not be included')
            stats.remove('sat')
        if 'idim' in stats and not _check_stat_available(npy_files, 'intrinsic-dimensionality'):
            warnings.warn('No .npy file was found for intrinsic dimensionality, stat will not be included')
            stats.remove('idim')
        result = _obtain_stats_from_npy(npy_files)
    else:
        if not _check_stat_available(npy_files, 'covariance'):
            raise FileNotFoundError("Covariance Matrix was not found, recomputation impossible")
        result = _recompute_stats(npy_files, stats, thresh)
    result.to_csv(savefile, sep=';')
    return result



import datetime

import keras
import tensorflow as tf

from keras import backend as K

import numpy as np
from numpy.linalg import LinAlgError


def get_preactivation_tensors(layers):
    """Get all valid layers for computing saturation."""
    dense_outputs = []
    for layer in layers:
        if 'dense' in layer.name:
            if '_input' in layer.name:
                # HACK
                continue
            # Get pre-activation
            if hasattr(layer, 'activation') and layer.activation.__name__ != \
                    'linear':
                preactivation_tensor = layer.output.op.inputs[0]
            else:
                preactivation_tensor = layer.output
            dense_outputs.append(preactivation_tensor)
    return dense_outputs


def initialize_preactivation_states(dense_outputs, obj):
    """Creates lists for `preactivation_states` dictionary."""
    for tensor in dense_outputs:
        layer_name = tensor.name.split('/')[0]
        obj.preactivation_states[layer_name] = []


def record_saturation(layers: str,
                      obj,
                      epoch: int,
                      logs: dict,
                      write_summary: bool = True):
    """Records saturation for layers into logs and writes summaries."""
    for layer in layers:
        layer_history = obj.preactivation_states[layer]
        if len(layer_history) < 2:  # ?
            continue
        history = np.stack(
            layer_history)[:, 0, :]  # get first representation of each batch
        history_T = history.T
        try:
            cov = np.cov(history_T)
        except LinAlgError:
            continue
        eig_vals, eig_vecs = np.linalg.eigh(cov)

        # Make a list of (eigenvalue, eigenvector) tuples
        eig_pairs = [(np.abs(eig_vals[i]), eig_vecs[:, i])
                     for i in range(len(eig_vals))]
        # Sort the (eigenvalue, eigenvector) tuples from high to low
        eig_pairs = sorted(eig_pairs, key=lambda x: x[0], reverse=True)
        eig_vals, eig_vecs = zip(*eig_pairs)
        tot = sum(eig_vals)

        # Get explained variance
        var_exp = [(i / tot) for i in eig_vals]

        # Get Simpson-diversity-index-based saturation
        weighted_sum = sum([x**2 for x in var_exp])  #
        logs[layer] = weighted_sum
        if write_summary:
            tf.summary.scalar(layer,
                              weighted_sum,
                              collections=['preactivation_state'])
    return logs


def get_layer_outputs(obj):
    """Get intermediate outputs aka. preactivation states."""
    layers = obj.model.layers[1:]
    dense_outputs = get_preactivation_tensors(layers)
    return dense_outputs


def save_intermediate_outputs(dense_outputs, obj):
    """Save outputs to obj."""
    for tensor in dense_outputs:
        layer_name = tensor.name.split('/')[0]

        # Route intermediate output, aka. preactivation state
        func = K.function([obj.model.input] + [K.learning_phase()], [tensor])
        intermediate_output = func([obj.input_data, 0.])[0]  # batch_nr x width

        obj.preactivation_states[layer_name].append(intermediate_output)


class CustomTensorBoard(tf.keras.callbacks.TensorBoard):
    """Extends the TensorBoard callback to allow adding custom summaries.
      From https://groups.google.com/forum/#!topic/keras-users/rEJ1xYqD3AM.

    Args:
        user_defined_freq: frequency (in epochs) at which to compute summaries
            defined by the user by calling tf.summary in the model code. If set to
            0, user-defined summaries won't be computed. Validation data must be
            specified for summary visualization.
        kwargs: Passed to tf.keras.callbacks.TensorBoard.
    """

    def __init__(self, user_defined_freq=0, **kwargs):
        self.user_defined_freq = user_defined_freq
        super(CustomTensorBoard, self).__init__(**kwargs)

    def on_epoch_begin(self, epoch, logs=None):
        """Add user-def. op to Model eval_function callbacks, reset batch count."""

        # check if histogram summary should be run for this epoch
        if self.user_defined_freq and epoch % self.user_defined_freq == 0:
            self._epoch = epoch
            # pylint: disable=protected-access
            # add the user-defined summary ops if it should run this epoch
            self.model._make_eval_function()
            if self.merged not in self.model._eval_function.fetches:
                self.model._eval_function.fetches.append(self.merged)
                self.model._eval_function.fetch_callbacks[
                    self.merged] = self._fetch_callback
            # pylint: enable=protected-access

        super(CustomTensorBoard, self).on_epoch_begin(epoch, logs=None)


    def on_epoch_end(self, epoch, logs=None):
        """Checks if summary ops should run next epoch, logs scalar summaries."""

        # pop the user-defined summary op after each epoch
        if self.user_defined_freq:
            # pylint: disable=protected-access
            if self.merged in self.model._eval_function.fetches:
                self.model._eval_function.fetches.remove(self.merged)
            if self.merged in self.model._eval_function.fetch_callbacks:
                self.model._eval_function.fetch_callbacks.pop(self.merged)
            # pylint: enable=protected-access

        super(CustomTensorBoard, self).on_epoch_end(epoch, logs=logs)


class SaturationMetric(keras.callbacks.Callback):
    """Keras callback for computing and logging layer saturation.

        Args:
            model: Keras model
            input_data: sample input to calculate layer saturation with, eg train
            print_freq
    """

    def __init__(self, model, input_data, print_freq=1):
        self.model = model
        self.input_data = input_data
        self.print_freq = print_freq

    def on_train_begin(self, logs=None):
        self.preactivation_states = {}
        layers = self.model.layers
        dense_outputs = get_preactivation_tensors(layers)
        initialize_preactivation_states(dense_outputs, self)

    def on_batch_end(self, batch, logs):
        if batch % 10 == 0:
            # TODO Check if has activation
            dense_outputs = get_layer_outputs(self)
            save_intermediate_outputs(dense_outputs, self)

    def on_epoch_end(self, epoch, logs):
        layers = self.preactivation_states.keys()
        logs = record_saturation(layers, self, epoch, logs)
        if epoch > 2:
            for layer in layers:
                try:
                    print("epoch = %4d  layer = %r  sat = %0.2f%%" \
                          % (epoch, layer, logs[layer]))
                    logs[layer] = self.preactivation_states[layer]
                except Exception as e:
                    print(e)


class SaturationLogger(keras.callbacks.Callback):
    """Keras callback for computing and logging layer saturation.

        Args:
            model: Keras model
            input_data: sample input to calculate layer saturation with, eg train
            print_freq
    """

    def __init__(self, model, input_data, print_freq=1):
        self.model = model
        self.input_data = input_data
        self.print_freq = print_freq

    def on_train_begin(self, logs=None):
        self.preactivation_states = {}
        layers = self.model.layers
        dense_outputs = get_preactivation_tensors(layers)
        initialize_preactivation_states(dense_outputs, self)

    def on_batch_end(self, batch, logs):
        if batch % 10 == 0:
            # TODO Check if has activation
            dense_outputs = get_layer_outputs(self)
            save_intermediate_outputs(dense_outputs, self)

    def on_epoch_end(self, epoch, logs):
        layers = self.preactivation_states.keys()
        logs = record_saturation(layers,
                                 self,
                                 epoch,
                                 logs,
                                 write_summary=False)
        if epoch > 2:
            # TODO: Integrate with Keras logging
            print_str = ""
            for layer in layers:
                print_str += "{:^10}: %{:4.2f} |".format(layer, logs[layer])
            print(print_str)



import logging
from typing import List, Dict, Any, Optional, Union
import torch
from collections import OrderedDict
from itertools import product
from torch.nn.modules.activation import ReLU
from torch.nn.modules.conv import Conv2d
from torch.nn.modules.linear import Linear
from torch.nn.modules import LSTM
#from mdp.utils import CovarianceMatrix
from delve.torch_utils import TorchCovarianceMatrix
from delve.writers import CompositWriter, NPYWriter
from delve.metrics import *
import delve
import warnings

logging.basicConfig(format='%(levelname)s:delve:%(message)s',
                    level=logging.INFO)


class CheckLayerSat(object):
    """Takes PyTorch layers, layer or model as `modules` and writes tensorboardX
    summaries to `logging_dir`. Outputs layer saturation with call to self.saturation().

    Args:
        logging_dir (str)  : destination for summaries
        modules (torch modules or list of modules) : layer-containing object
        log_interval (int) : steps between writing summaries
        stats (list of str): list of stats to collect

            supported stats are:
                lsat       : layer saturation

        sat_method         : How to calculate saturation

            Choice of:

                cumvar99   : Proportion of eigenvalues needed to explain 99% of variance
                simpson_di : Simpson diversity index (weighted sum) of explained variance
                             ratios of eigenvalues
                all        : All available methods are logged

        include_conv       : setting to False includes only linear layers
        conv_method        : how to subsample convolutional layers
        verbose (bool)     : print saturation for every layer during training

    """

    def __init__(
            self,
            savefile: str,
            save_to: Union[str, delve.writers.AbstractWriter],
            modules: torch.nn.Module,
            writer_args: Optional[Dict[str, Any]] = None,
            log_interval=1,
            max_samples=None,
            stats: list = ['lsat'],
            layerwise_sat: bool = True,
            reset_covariance: bool = True,
            average_sat: bool = False,
            ignore_layer_names: List[str] = [],
            include_conv: bool = True,
            conv_method: str = 'channelwise',
            sat_threshold: str = .99,
            verbose=False,
            device='cuda:0',
    ):
        self.verbose = verbose
        self.include_conv = include_conv
        self.conv_method = conv_method
        self.threshold = sat_threshold
        self.layers = self.get_layers_recursive(modules)
        self.max_samples = max_samples
        self.log_interval = log_interval
        self.reset_covariance = reset_covariance

        if writer_args is None:
            writer_args = {}
        writer_args['savepath'] = savefile

        self.writer = self._get_writer(save_to, writer_args)
        self.interval = log_interval

        self._warn_if_covariance_not_saveable(stats)

        self.logs, self.stats = self._check_stats(stats)
        self.layerwise_sat = layerwise_sat
        self.average_sat = average_sat
        self.ignore_layer_names = ignore_layer_names
        self.seen_samples = {
            'train': {},
            'eval': {}
        }
        self.global_steps = 0
        self.global_hooks_registered = False
        self.is_notebook = None
        self.device = device
        for name, layer in self.layers.items():
            if isinstance(layer, Conv2d) or isinstance(layer, Linear) or isinstance(layer, LSTM):
                self._register_hooks(layer=layer,
                                     layer_name=name,
                                     interval=log_interval)

    def _warn_if_covariance_not_saveable(self, stats: List[str]):
        warn = False
        if 'cov' in stats:
            if isinstance(self.writer, CompositWriter):
                for writer in self.writer.writers:
                    if isinstance(writer, NPYWriter):
                        return
                warn = True
            elif not isinstance(self.writer, NPYWriter):
                warn = True
        if warn:
            warnings.warn("'cov' was selected as stat, but 'npy' (NPYWriter) is not used as a save strategy,"
                          "which is the only writer able to save the covariance matrix. The training and logging will"
                          "run normally, but the covariance matrix will not be saved. Note that you can add multiple writers"
                          "by passing a list.")


    def __getattr__(self, name):
        if name.startswith('add_') and name != 'add_saturations':
            return getattr(self.writer, name)
        else:
            try:
                # Redirect to writer object
                return self.writer.__getattribute__(name)
            except:
                # Default behaviour
                return self.__getattribute__(name)

    def __repr__(self):
        return self.layers.keys().__repr__()

    def close(self):
        """User endpoint to close writer and progress bars."""
        return self.writer.close()

    def _format_saturation(self, saturation_status):
        raise NotImplementedError

    def _check_stats(self, stats: list):
        if not isinstance(stats, list):
            stats = list(stats)
        supported_stats = [
            'lsat',
            'idim',
            'cov',
        ]
        compatible = [stat in supported_stats for stat in stats]
        incompatible = [i for i, x in enumerate(compatible) if not x]
        assert all(compatible), "Stat {} is not supported".format(
            stats[incompatible[0]])

        name_mapper = {
            'lsat': 'saturation',
            'idim': 'intrinsic-dimensionality',
            'cov': 'covariance-matrix'
        }

        logs = {
            f'{mode}-{name_mapper[stat]}': OrderedDict()
            for mode, stat in product(['train', 'eval'], stats)
        }

        return logs, stats

    def _add_conv_layer(self, layer: torch.nn.Module):
        layer.out_features = layer.out_channels
        layer.conv_method = self.conv_method

    def get_layer_from_submodule(self, submodule: torch.nn.Module, layers: dict, name_prefix: str = ''):
            if len(submodule._modules) > 0:
                for idx, (name, subsubmodule) in enumerate(submodule._modules.items()):
                    new_prefix = name if name_prefix == '' else name_prefix+'-'+name
                    self.get_layer_from_submodule(subsubmodule, layers, new_prefix)
                return layers
            else:
                layer_name = name_prefix
                layer_type = layer_name
                if not isinstance(submodule, Conv2d) and not isinstance(submodule, Linear) and not isinstance(submodule, LSTM):
                    print(f"Skipping {layer_type}")
                    return layers
                if isinstance(submodule, Conv2d) and self.include_conv:
                    self._add_conv_layer(submodule)
                layers[layer_name] = submodule
                print('added layer {}'.format(layer_name))
                return layers

    def get_layers_recursive(self, modules: Union[list, torch.nn.Module]):
        layers = {}
        if not isinstance(modules, list) and not hasattr(
                modules, 'out_features'):
            # is a model with layers
            # check if submodule
            submodules = modules._modules  # OrderedDict
            layers = self.get_layer_from_submodule(modules, layers, '')
        else:
            for module in modules:
                layers = self.get_layer_from_submodule(module, layers, '')
        return layers

    def _get_writer(self, save_to, writers_args) -> delve.writers.AbstractWriter:
        """Create a writer to log history to `writer_dir`."""
        if issubclass(type(save_to), delve.writers.AbstractWriter):
            return save_to
        if isinstance(save_to, list):
            all_writers = []
            for saver in save_to:
                all_writers.append(self._get_writer(save_to=saver, writers_args=writers_args))
            return CompositWriter(all_writers)
        if hasattr(delve, save_to):
            writer = getattr(delve, save_to)(**writers_args)
        else:
            raise ValueError('Illegal argument for save_to "{}"'.format(save_to))
        return writer

    def _register_hooks(self, layer: torch.nn.Module, layer_name: str,
                        interval):
        if not hasattr(layer, 'eval_layer_history'):
            layer.eval_layer_history = []
        if not hasattr(layer, 'train_layer_history'):
            layer.train_layer_history = []
        if not hasattr(layer, 'layer_svd'):
            layer.layer_svd = None
        if not hasattr(layer, 'forward_iter'):
            layer.forward_iter = 0
        if not hasattr(layer, 'interval'):
            layer.interval = interval
        if not hasattr(layer, 'writer'):
            layer.writer = self.writer
        if not hasattr(layer, 'name'):
            layer.name = layer_name
        self.register_forward_hooks(layer, self.stats)
        return self

    def _record_stat(self, activations_batch: torch.Tensor, lstm_ae: bool, layer: torch.nn.Module, training_state: str, stat: str):
        if activations_batch.dim() == 4:  # conv layer (B x C x H x W)
            if self.conv_method == 'median':
                shape = activations_batch.shape
                reshaped_batch = activations_batch.reshape(shape[0], shape[1], shape[2] * shape[3])
                activations_batch, _ = torch.median(reshaped_batch, dim=2)  # channel median
            elif self.conv_method == 'max':
                shape = activations_batch.shape
                reshaped_batch = activations_batch.reshape(shape[0], shape[1], shape[2] * shape[3])
                activations_batch, _ = torch.max(reshaped_batch, dim=2)  # channel median
            elif self.conv_method == 'mean':
                activations_batch = torch.mean(activations_batch, dim=(2, 3))
            elif self.conv_method == 'flatten':
                activations_batch = activations_batch.view(activations_batch.size(0), -1)
            elif self.conv_method == 'channelwise':
                reshaped_batch: torch.Tensor = activations_batch.permute([1, 0, 2, 3])
                shape = reshaped_batch.shape
                reshaped_batch: torch.Tensor = reshaped_batch.flatten(1)
                reshaped_batch: torch.Tensor = reshaped_batch.permute([1, 0])
                activations_batch = reshaped_batch

        if layer.name in self.logs[f'{training_state}-{stat}']:
            self.logs[f'{training_state}-{stat}'][layer.name].update(activations_batch, lstm_ae)
        else:
            self.logs[f'{training_state}-{stat}'][layer.name] = TorchCovarianceMatrix(device=self.device)
            self.logs[f'{training_state}-{stat}'][layer.name].update(activations_batch, lstm_ae)

    def register_forward_hooks(self, layer: torch.nn.Module, stats: list):
        """Register hook to show `stats` in `layer`."""

        def record_layer_saturation(layer: torch.nn.Module, input, output):
            """Hook to register in `layer` module."""

            # Increment step counter
            layer.forward_iter += 1
            
            # VAE output is a tuple; Hence output.data throw exception
            lstm_ae = False
            if layer.name in ['encoder_lstm','encoder_output','decoder_lstm','decoder_output']:
                output = output[1][0]
                lstm_ae = True

            training_state = 'train' if layer.training else 'eval'
            if not layer.name in self.seen_samples[training_state]:
                self.seen_samples[training_state][layer.name] = 0
            if (self.max_samples is None or self.seen_samples[training_state][layer.name] < self.max_samples) and layer.forward_iter % self.log_interval == 0:
                num_samples = min(output.data.shape[0], self.max_samples-self.seen_samples[training_state][layer.name]) if self.max_samples is not None else output.data.shape[0]
                activations_batch = output.data[:num_samples]
                self.seen_samples[training_state][layer.name] += num_samples
                if self.verbose:
                    print("seen {} samples on layer {}".format(self.seen_samples[training_state][layer.name], layer.name))

                eig_vals = None

                if 'lsat' in stats:
                    self._record_stat(activations_batch, lstm_ae, layer, training_state, 'saturation')
                if 'idim' in stats:
                    self._record_stat(activations_batch, lstm_ae, layer, training_state, 'intrinsic-dimensionality')
                if 'cov' in stats:
                    self._record_stat(activations_batch, lstm_ae, layer, training_state, 'covariance-matrix')

        layer.register_forward_hook(record_layer_saturation)

    def add_saturations(self, save=True):
        """
        Computes saturation and saves all stats
        :return:
        """
        for key in self.logs:
            train_sats = []
            val_sats = []
            if '-saturation' in key:
                for i, layer_name in enumerate(self.logs[key]):
                    if layer_name in self.ignore_layer_names:
                        continue
                    if self.logs[key][layer_name]._cov_mtx is None:
                        raise ValueError('Attempting to compute saturation when covariance is not initialized')
                    sat = compute_saturation(self.logs[key][layer_name].fix(), thresh=self.threshold)
                    self.seen_samples[key.split('-')[0]][layer_name] = 0
                    if self.reset_covariance:
                        self.logs[key][layer_name]._cov_mtx = None
                    if self.layerwise_sat:
                        name = key+'_'+layer_name
                        self.writer.add_scalar(name, sat)
                    if 'eval' in key:
                        val_sats.append(sat)
                    elif 'train' in key:
                        train_sats.append(sat)
            elif '-intrinsic' in key:
                for i, layer_name in enumerate(self.logs[key]):
                    if layer_name in self.ignore_layer_names:
                        continue
                    if self.logs[key][layer_name]._cov_mtx is None:
                        raise ValueError('Attempting to compute intrinsic dimensionality when covariance is not initialized')
                    intrinsic = compute_intrinsic_dimensionality(self.logs[key][layer_name].fix(), thresh=self.threshold)
                    self.seen_samples[key.split('-')[0]][layer_name] = 0
                    if self.reset_covariance:
                        self.logs[key][layer_name]._cov_mtx = None
                    if self.layerwise_sat:
                        name = key+'_'+layer_name
                        self.writer.add_scalar(name, intrinsic)
            elif '-covariance' in key:
                for i, layer_name in enumerate(self.logs[key]):
                    if layer_name in self.ignore_layer_names:
                        continue
                    if self.logs[key][layer_name]._cov_mtx is None:
                        raise ValueError('Attempting to compute intrinsic dimensionality when covariance is not initialized')
                    cov_mat = self.logs[key][layer_name].fix()
                    self.seen_samples[key.split('-')[0]][layer_name] = 0
                    if self.reset_covariance:
                        self.logs[key][layer_name]._cov_mtx = None
                    if self.layerwise_sat:
                        name = key+'_'+layer_name
                        self.writer.add_scalar(name, cov_mat)

        if self.average_sat:
            self.writer.add_scalar('average-train-sat', np.mean(train_sats))
            self.writer.add_scalar('average-eval-sat', np.mean(val_sats))

        if save:
            self.save()

    def save(self):
        self.writer.save()



try:
    import torch
    from delve.torchcallback import CheckLayerSat
    from delve.tools import reconstruct_csv_from_npy_data
except ImportError:
    pass
name = "delve"

from delve.writers import PrintWriter as console
from delve.writers import NPYWriter as npy
from delve.writers import CSVWriter as csv
from delve.writers import CSVandPlottingWriter as plot
from delve.writers import CSVandPlottingWriter as plotcsv

from delve.writers import CSVandPlottingWriter as csvplot
from delve.writers import TensorBoardWriter as tensorboard


import torch
from torch.nn import Module
import numpy as np

global num


def rvs(dim=3):
    random_state = np.random
    H = np.eye(dim)
    D = np.ones((dim,))
    for n in range(1, dim):
        x = random_state.normal(size=(dim - n + 1,))
        D[n - 1] = np.sign(x[0])
        x[0] -= D[n - 1] * np.sqrt((x * x).sum())
        # Householder transformation
        Hx = (np.eye(dim - n + 1) - 2. * np.outer(x, x) / (x * x).sum())
        mat = np.eye(dim)
        mat[n - 1:, n - 1:] = Hx
        H = np.dot(H, mat)
        # Fix the last sign such that the determinant is 1
    D[-1] = (-1) ** (1 - (dim % 2)) * D.prod()
    # Equivalent to np.dot(np.diag(D), H) but faster, apparently
    H = (D * H.T).T
    return H


def change_all_pca_layer_thresholds_and_inject_random_directions(threshold: float, network: Module,
                                                                 verbose: bool = False, device='cpu'):
    """
    This is a simple test function which changes the pca threshold of all PCA-Layers in a given network and
    injection random orthonormal bases instead of the base constructoed of eigenvectors.
    This function can be used to test the properties of PCA-Projections in a given network.
    """
    in_dims = []
    fs_dims = []
    sat = []
    for module in network.modules():
        if isinstance(module, LinearPCALayer):
            module.threshold = threshold
            fake_base = rvs(module.fs_dim)[:, :module.in_dim]
            in_dims.append(module.in_dim)
            fs_dims.append(module.fs_dim)
            sat.append(module.sat)
            fake_projection = fake_base @ fake_base.T
            module.transformation_matrix.data = torch.from_numpy(fake_projection.astype('float32')).to(device)
            if verbose:
                print(f'Changed threshold for layer {module} to {threshold}')
        elif isinstance(module, Conv2DPCALayer):
            module.threshold = threshold
            in_dims.append(module.in_dim)
            fs_dims.append(module.fs_dim)
            sat.append(module.sat)
            fake_base = rvs(module.fs_dim)[:, :module.in_dim]
            fake_projection = fake_base @ fake_base.T
            module.transformation_matrix.data = torch.from_numpy(fake_projection.astype('float32')).to(device)
            weight = torch.nn.Parameter(module.transformation_matrix.unsqueeze(2).unsqueeze(3))
            module.convolution.weight = weight
            if verbose:
                print(f'Changed threshold for layer {module} to {threshold}')
    return sat, in_dims, fs_dims


def change_all_pca_layer_thresholds(threshold: float, network: Module, verbose: bool = False):
    """
    Changes the Threshold of all pca layer in a given network to a specified value. Automatically recomputes eigenvalues
    """
    in_dims = []
    fs_dims = []
    sat = []
    for module in network.modules():
        if isinstance(module, Conv2DPCALayer) or isinstance(module, LinearPCALayer):
            module.threshold = threshold
            in_dims.append(module.in_dim)
            fs_dims.append(module.fs_dim)
            sat.append(module.sat)
            if verbose:
                print(f'Changed threshold for layer {module} to {threshold}')
    return sat, in_dims, fs_dims


class LinearPCALayer(Module):
    """
    The Linear PCA Layer Projects forces the latent representations into the approximated latent space at inference time.
    The linear PCA Layers is geared towards linear layer.
    """
    num = 0

    def __init__(self, in_features: int,
                 threshold: float = .99,
                 keepdim: bool = True,
                 verbose: bool = False,
                 gradient_epoch_start: int = 20,
                 centering: bool = True):
        super(LinearPCALayer, self).__init__()
        self.register_buffer('eigenvalues', torch.zeros(in_features, dtype=torch.float64))
        self.register_buffer('eigenvectors', torch.zeros((in_features, in_features), dtype=torch.float64))
        self.register_buffer('_threshold', torch.Tensor([threshold]).type(torch.float64))
        self.register_buffer('sum_squares', torch.zeros((in_features, in_features), dtype=torch.float64))
        self.register_buffer('seen_samples', torch.zeros(1, dtype=torch.float64))
        self.register_buffer('running_sum', torch.zeros(in_features, dtype=torch.float64))
        self.register_buffer('mean', torch.zeros(in_features, dtype=torch.float32))
        self.keepdim: bool = keepdim
        self.verbose: bool = verbose
        self.pca_computed: bool = True
        self.gradient_epoch = gradient_epoch_start
        self.epoch = 0
        self.name = f'pca{LinearPCALayer.num}'
        LinearPCALayer.num += 1
        self._centering = centering

    @property
    def threshold(self) -> float:
        return self._threshold

    @threshold.setter
    def threshold(self, threshold: float) -> None:
        self._threshold.data = torch.Tensor([threshold]).type(torch.float64).to(self.threshold.device)
        self._compute_pca_matrix()

    @property
    def centering(self):
        return self._centering

    @centering.setter
    def centering(self, centring: bool):
        self._centering = centring
        self._compute_pca_matrix()

    def _update_autorcorrelation(self, x: torch.Tensor) -> None:
        x = x.type(torch.float64)
        self.sum_squares.data += torch.matmul(x.transpose(0, 1), x)
        self.running_sum += x.sum(dim=0)
        self.seen_samples.data += x.shape[0]

    def _compute_autorcorrelation(self) -> torch.Tensor:
        tlen = self.seen_samples
        cov_mtx = self.sum_squares
        avg = self.running_sum / tlen
        if self.centering:
            avg_mtx = torch.ger(avg, avg)
            cov_mtx = cov_mtx - avg_mtx

            cov_mtx = cov_mtx / tlen
        self.mean.data = avg.type(torch.float32)

        # np.save(self.name+'_cov_mtx.npy', cov_mtx.cpu().numpy())
        # np.save(self.name+'_mean.npy', self.mean.cpu().numpy())
        return cov_mtx

    def _compute_eigenspace(self):
        self.eigenvalues.data, self.eigenvectors.data = self._compute_autorcorrelation().symeig(True)
        self.eigenvalues.data, idx = self.eigenvalues.sort(descending=True)
        # correct numerical error, matrix must be positivly semi-definitie
        self.eigenvalues[self.eigenvalues < 0] = 0
        self.eigenvectors.data = self.eigenvectors[:, idx]

    def _reset_autorcorrelation(self):
        self.sum_squares.data = torch.zeros(self.sum_squares.shape, dtype=torch.float64).to(self.sum_squares.device)
        self.seen_samples.data = torch.zeros(self.seen_samples.shape, dtype=torch.float64).to(self.sum_squares.device)
        self.running_sum.data = torch.zeros(self.running_sum.shape, dtype=torch.float64).to(self.sum_squares.device)

    def _compute_pca_matrix(self):
        if self.verbose:
            print('computing autorcorrelation for Linear')
            # print('Mean pre-activation vector:', self.mean)
        percentages = self.eigenvalues.cumsum(0) / self.eigenvalues.sum()
        eigen_space = self.eigenvectors[:, percentages < self.threshold]
        if eigen_space.shape[1] == 0:
            eigen_space = self.eigenvectors[:, :1]
            print(f'Detected singularity defaulting to single dimension {eigen_space.shape}')
        elif self.threshold - (percentages[percentages < self.threshold][-1]) > 0.02:
            print(f'Highest cumvar99 is {percentages[percentages < self.threshold][-1]}, extending eigenspace by one dimension for eigenspace of {eigen_space.shape}')
            eigen_space = self.eigenvectors[:, :eigen_space.shape[1] + 1]

        sat = round((eigen_space.shape[1] / self.eigenvalues.shape[0]) * 100, 4)
        fs_dim = eigen_space.shape[0]
        in_dim = eigen_space.shape[1]
        if self.verbose:
            print(f'Saturation: {round(eigen_space.shape[1] / self.eigenvalues.shape[0], 4)}%', 'Eigenspace has shape',
                  eigen_space.shape)
        self.transformation_matrix: torch.Tensor = eigen_space.matmul(eigen_space.t()).type(torch.float32)
        self.reduced_transformation_matrix: torch.Tensor = eigen_space.type(torch.float32)
        self.sat, self.in_dim, self.fs_dim = sat, in_dim, fs_dim

    def forward(self, x):
        if self.training:
            self.pca_computed = False
            self._update_autorcorrelation(x)
            return x
        else:
            if not self.pca_computed:
                self._compute_autorcorrelation()
                self._compute_eigenspace()
                self._compute_pca_matrix()
                self.pca_computed = True
                self._reset_autorcorrelation()
                self.epoch += 1
            if self.keepdim:
                if not self.centering:
                    return x @ self.transformation_matrix.t()
                else:
                    return ((x - self.mean) @ self.transformation_matrix.t()) + self.mean
            else:
                if not self.centering:
                    return x @ self.reduced_transformation_matrix
                else:
                    return ((x - self.mean) @ self.reduced_transformation_matrix) + self.mean


class Conv2DPCALayer(LinearPCALayer):
    """
    The Convolutional PCA Layer Projects forces the latent representations into the approximated latent space at inference time.
    The Convolutional PCA Layers is geared towards convolutional layers. Each filter for that matter is treated as a latent space dimension.
    Every position on that filter-stack is treated as an individual observeration.
    """

    def __init__(self, in_filters, threshold: float = 0.99, verbose: bool = True, gradient_epoch_start: int = 20,
                 centering: bool = False):
        super(Conv2DPCALayer, self).__init__(centering=centering, in_features=in_filters, threshold=threshold,
                                             keepdim=True, verbose=verbose, gradient_epoch_start=gradient_epoch_start)
        if verbose:
            print('Added Conv2D PCA Layer')
        self.convolution = torch.nn.Conv2d(in_channels=in_filters,
                                           out_channels=in_filters,
                                           kernel_size=1, stride=1, bias=True)
        self.mean_subtracting_convolution = torch.nn.Conv2d(in_channels=in_filters,
                                                            out_channels=in_filters,
                                                            kernel_size=1, stride=1, bias=True)
        self.mean_subtracting_convolution.weight = torch.nn.Parameter(
            torch.zeros((in_filters, in_filters)).unsqueeze(2).unsqueeze(3)
        )

    def _compute_pca_matrix(self):
        if self.verbose:
            print('computing autorcorrelation for Conv2D')
        super()._compute_pca_matrix()
        # unsequeeze the matrix into 1x1xDxD in order to make it behave like a 1x1 convolution
        weight = torch.nn.Parameter(
            self.transformation_matrix.unsqueeze(2).unsqueeze(3)
        )
        self.convolution.weight = weight

        self.mean_subtracting_convolution.weight = torch.nn.Parameter(
            torch.zeros_like(self.transformation_matrix).unsqueeze(2).unsqueeze(3)
        )

        if self.centering:
            self.convolution.bias = torch.nn.Parameter(
                self.mean.type(torch.float32)
            )
            self.mean_subtracting_convolution.bias = torch.nn.Parameter(
                -self.mean.type(torch.float32)
            )
        else:
            self.convolution.bias = torch.nn.Parameter(
                torch.zeros_like(self.mean)
            )
            self.mean_subtracting_convolution.bias = torch.nn.Parameter(
                torch.zeros_like(self.mean)
            )

    def forward(self, x):
        if self.training:
            self.pca_computed = False
            swapped: torch.Tensor = x.permute([1, 0, 2, 3])
            flattened: torch.Tensor = swapped.flatten(1)
            reshaped_batch: torch.Tensor = flattened.permute([1, 0])
            self._update_autorcorrelation(reshaped_batch)
            return x
        else:
            if not self.pca_computed:
                self._compute_autorcorrelation()
                self._compute_eigenspace()
                self._compute_pca_matrix()
                self._reset_autorcorrelation()
                self.pca_computed = True
            # if self.centering:
            x1 = self.mean_subtracting_convolution(x)
            x = x + x1
            return self.convolution(x)



from typing import Dict, Union

import numpy as np
import torch


def compute_intrinsic_dimensionality(cov: torch.Tensor, thresh: float = .99) -> int:
    """
    Compute the intrinsic dimensionality based on the covariance matrix
    :param cov: the covariance matrix as a torch tensor
    :param thresh: the delta value for the explained variance of the covariance matrix
    :return: The intrinsic dimensionality, which is a single integer value greater than 0
    """
    eig_vals, eigen_space = cov.symeig(True)
    eig_vals, idx = eig_vals.sort(descending=True)
    eig_vals[eig_vals < 0] = 0
    percentages = eig_vals.cumsum(0) / eig_vals.sum()
    eigen_space = eigen_space[:, percentages < thresh]
    if eigen_space.shape[1] == 0:
        eigen_space = eigen_space[:, :1]
    elif thresh - (percentages[percentages < thresh][-1]) > 0.02:
        eigen_space = eigen_space[:, :eigen_space.shape[1] + 1]
    return eigen_space.shape[1]


def compute_saturation(cov: torch.Tensor, thresh: float = .99) -> float:
    """
    Computes the saturation
    :param cov: the covariance matrix as a torch tensor
    :param thresh: the delta value for the explained variance of the covariance matrix
    :return: a value between 0 and 1
    """
    intrinsic_dimensionality = compute_intrinsic_dimensionality(cov, thresh)
    feature_space_dimensionality = cov.shape[0]

    return intrinsic_dimensionality / feature_space_dimensionality



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import os
import os.path as path
from delve.version import __version__
# Get package version
exec(open('delve/version.py', 'r').read())

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['tensorboardX', 'tqdm', 'MDP']

this_dir = path.abspath(path.dirname(__file__))
with open(os.path.join(this_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='delve',
    version=__version__,
    description=
    'Delve lets you monitor PyTorch model layer saturation during training',
    url='https://github.com/delve-team/delve',
    author='Justin Shenk',
    author_email='shenk.justin@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT license',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='!= 3.0.*, != 3.1.*',
    packages=find_packages(),
    include_package_data=True,
    keywords='deep learning layer saturation pruning spectral tensorboard network',
    zip_safe=False,
)
